from ._dock_widgets import train_from_viewer, segment_data, combine_layers,  \
    assess_segmentation, compare_segmentations

__all__ = [
    'train_from_viewer', 
    'segment_data', 
    'combine_layers', 
    'generate_ground_truth', 
    'assess_segmentation', 
    'compare_segmentations'
]

__version__ = "0.0.1"

